package it.unibs.fp.patenteelibbrettopregoh;

import javax.xml.stream.*;
import java.io.*;
import java.lang.reflect.Type;
import java.util.ArrayList;

import com.google.gson.Gson;
import com.google.gson.JsonIOException;
import com.google.gson.JsonSyntaxException;
import com.google.gson.reflect.TypeToken;

import java.io.FileWriter;

/**
 * Parser.java
 * Questa classe è responsabile della lettura e scrittura dei dati di gioco in formato JSON e XML.
 * Utilizza la libreria Gson per la serializzazione e deserializzazione dei dati.
 */
public class Parser {
    /**
     * Metodo per il parsing delle persone da un file JSON o XML.
     *
     * @param formato  il formato del file (JSON o XML)
     * @param fileName il nome del file senza estensione
     * @return una lista di oggetti Persona
     */
    public static ArrayList<Persona> parsePersone(Formato formato, String fileName) {
        ArrayList<Persona> persone = new ArrayList<>();
        if (formato == Formato.XML) {
            parsePersoneXML(fileName, persone);

        }
        else {
            persone = parsePersoneJSON(fileName, persone);

        }
        return persone;
    }

    /**
     * Metodo per il parsing delle persone da un file JSON.
     *
     * @param fileName il nome del file senza estensione
     * @param persone  la lista di persone da riempire
     * @return la lista di persone
     */
    private static ArrayList<Persona> parsePersoneJSON(String fileName, ArrayList<Persona> persone) {
        fileName = fileName + ".json";
        try (FileReader reader = new FileReader(fileName)) {
            Gson gson = new Gson();
            Type listType = new TypeToken<ArrayList<Persona>>(){}.getType();
            persone = gson.fromJson(reader, listType);
        } catch (JsonIOException | JsonSyntaxException | IOException e) {
            System.out.println("Error reading the JSON file: " + e.getMessage());
        }
        return persone;
    }

    /**
     * Metodo per il parsing delle persone da un file XML.
     *
     * @param fileName il nome del file senza estensione
     * @param persone  la lista di persone da riempire
     */
    private static void parsePersoneXML(String fileName, ArrayList<Persona> persone) {
        fileName = fileName + ".xml";
        XMLInputFactory xmlif;
        XMLStreamReader xmlr;
        String currentElement = null;
        String nome = null;
        String cognome = null;
        String sesso = null;
        String dataNascita = null;
        String comuneNascita = null;
        String codiceFiscale = null;
        String scadenzaDocumento = null;

        try (FileInputStream reader = new FileInputStream(fileName)) {
            xmlif = XMLInputFactory.newInstance();
            xmlr = xmlif.createXMLStreamReader(fileName, reader);

            while (xmlr.hasNext()) {
                int event = xmlr.getEventType();
                switch (event) {
                    case XMLStreamConstants.START_ELEMENT:
                        currentElement = xmlr.getLocalName(); // <- sposta qui, nel posto giusto
                        break;
                    case XMLStreamConstants.CHARACTERS:
                        String text = xmlr.getText().trim();
                        if (!text.isEmpty() && currentElement != null) {
                            switch (currentElement) {
                                case "nome" -> nome = text;
                                case "cognome" -> cognome = text;
                                case "sesso" -> sesso = text;
                                case "data_nascita" -> dataNascita = text;
                                case "comune_nascita" -> comuneNascita = text;
                                case "codice_fiscale" -> codiceFiscale = text;
                                case "scadenza_documento" -> scadenzaDocumento = text;
                            }
                        }
                        break;
                    case XMLStreamConstants.END_ELEMENT:
                        String endTag = xmlr.getLocalName();
                        if (endTag.equals("persona")) {
                            Persona persona = new Persona(nome, cognome, sesso, dataNascita, comuneNascita, codiceFiscale, scadenzaDocumento);
                            persone.add(persona);
                        }
                        currentElement = null; // reset solo alla fine di un elemento
                        break;
                }
                xmlr.next(); // importante: avanza solo una volta per ciclo
            }
            xmlr.close();
        } catch (FactoryConfigurationError | XMLStreamException | IOException e) {
            System.out.println("Error reading the file: " + e.getMessage());
        }
    }

    /**
     * Metodo per il parsing dei comuni da un file JSON o XML.
     *
     * @param formato  il formato del file (JSON o XML)
     * @param fileName il nome del file senza estensione
     * @return un oggetto Comuni contenente la lista di comuni
     */
    public static Comuni parseComuni(Formato formato, String fileName) {
        Comuni comuni = new Comuni();
        if (formato == Formato.XML) {
            parseComuniXML(fileName, comuni);
        }
        else {
            comuni = parseComuniJSON(fileName, comuni);
        }

        return comuni;
    }

    /**
     * Metodo per il parsing dei comuni da un file JSON.
     *
     * @param fileName il nome del file senza estensione
     * @param comuni   l'oggetto Comuni da riempire
     * @return l'oggetto Comuni
     */
    private static Comuni parseComuniJSON(String fileName, Comuni comuni) {
        fileName = fileName + ".json";
        try (FileReader reader = new FileReader(fileName)) {
            Gson gson = new Gson();
            Type listType = new TypeToken<ArrayList<Comune>>() {}.getType();
            ArrayList<Comune> listaComuni = gson.fromJson(reader, listType);

            comuni = new Comuni();
            for (Comune comune : listaComuni) {
                comuni.addComune(comune);
            }

        } catch (JsonIOException | JsonSyntaxException | IOException e) {
            System.out.println("Error reading the JSON file: " + e.getMessage());
        }
        return comuni;
    }

    /**
     * Metodo per il parsing dei comuni da un file XML.
     *
     * @param fileName il nome del file senza estensione
     * @param comuni   l'oggetto Comuni da riempire
     */
    private static void parseComuniXML(String fileName, Comuni comuni) {
        fileName = fileName + ".xml";
        XMLInputFactory xmlif;
        XMLStreamReader xmlr;
        String currentElement = null;
        String nome = null;
        String codice = null;

        try (FileInputStream reader = new FileInputStream(fileName)) {
            xmlif = XMLInputFactory.newInstance();
            xmlr = xmlif.createXMLStreamReader(fileName, reader);

            while (xmlr.hasNext()) {
                int event = xmlr.getEventType();

                switch (event) {

                    case XMLStreamConstants.START_ELEMENT:
                        currentElement = xmlr.getLocalName();
                        break;


                    case XMLStreamConstants.CHARACTERS:
                        String text = xmlr.getText().trim();
                        if (!text.isEmpty() && currentElement != null) {
                            if (currentElement.equals("nome")) {
                                nome = text;
                            } else if (currentElement.equals("codice")) {
                                codice = text;
                            }
                        }
                        break;

                    case XMLStreamConstants.END_ELEMENT:
                        String endTag = xmlr.getLocalName();
                        if (endTag.equals("comune")) {
                            Comune comune = new Comune(nome, codice);
                            comuni.addComune(comune);

                            nome = null;
                            codice = null;
                        }
                        currentElement = null;
                        break;
                }

                xmlr.next();
            }

            xmlr.close();
        } catch (FactoryConfigurationError | XMLStreamException | IOException e) {
            System.out.println("Error reading the file: " + e.getMessage());
        }
    }

    /**
     * Metodo per il salvataggio del diario in formato JSON o XML.
     *
     * @param fileName      il nome del file senza estensione
     * @param pagineDiario  la lista di pagine del diario
     * @param formato       il formato del file (JSON o XML)
     */
    public static void saveDiario (String fileName, ArrayList<PaginaDiario> pagineDiario, Formato formato) {
        if (formato == Formato.XML) {
            salvaDiarioXML(fileName, pagineDiario);

        } else if (formato == Formato.JSON) {
            salvaDiarioJSON(fileName, pagineDiario);
        }
    }

    /**
     * Metodo per il salvataggio del diario in formato JSON.
     *
     * @param fileName      il nome del file senza estensione
     * @param pagineDiario  la lista di pagine del diario
     */
    private static void salvaDiarioJSON(String fileName, ArrayList<PaginaDiario> pagineDiario) {
        fileName = fileName + ".json";
        Gson gson = new Gson();
        try (FileWriter writer = new FileWriter(fileName)) {
            gson.toJson(pagineDiario, writer);
        } catch (JsonIOException | IOException e) {
            System.out.println("Error in initializing the writer:");
            System.out.println(e.getMessage());
        }
    }

    /**
     * Metodo per il salvataggio del diario in formato XML.
     *
     * @param fileName      il nome del file senza estensione
     * @param pagineDiario  la lista di pagine del diario
     */
    private static void salvaDiarioXML(String fileName, ArrayList<PaginaDiario> pagineDiario) {
        fileName = fileName + ".xml";
        try (FileOutputStream outputStream = new FileOutputStream(fileName)) {
            XMLOutputFactory xmlOutputFactory = XMLOutputFactory.newInstance();
            XMLStreamWriter xmlWriter = xmlOutputFactory.createXMLStreamWriter(outputStream, "UTF-8");

            xmlWriter.writeStartDocument("UTF-8", "1.0");
            xmlWriter.writeStartElement("diario");

            for (PaginaDiario pagina : pagineDiario) {
                xmlWriter.writeStartElement("giornata");
                xmlWriter.writeAttribute("numero", String.valueOf(pagina.getGiorno()));

                xmlWriter.writeStartElement("data");
                xmlWriter.writeCharacters(pagina.getData());
                xmlWriter.writeEndElement();

                xmlWriter.writeStartElement("bilancio");
                xmlWriter.writeCharacters(String.valueOf(pagina.getBilancio()));
                xmlWriter.writeEndElement();

                xmlWriter.writeStartElement("persone");
                xmlWriter.writeAttribute("numero", String.valueOf(pagina.getPersoneControllate().size()));

                for (int i = 0; i < pagina.getPersoneControllate().size(); i++) {
                    PersonaControllata persona = pagina.getPersoneControllate().get(i);

                    xmlWriter.writeStartElement("persona");
                    xmlWriter.writeAttribute("id", String.valueOf(i + 1));

                    xmlWriter.writeStartElement("nome");
                    xmlWriter.writeCharacters(persona.getNome());
                    xmlWriter.writeEndElement();

                    xmlWriter.writeStartElement("cognome");
                    xmlWriter.writeCharacters(persona.getCognome());
                    xmlWriter.writeEndElement();

                    xmlWriter.writeStartElement("sesso");
                    xmlWriter.writeCharacters(persona.getSesso());
                    xmlWriter.writeEndElement();

                    xmlWriter.writeStartElement("data_nascita");
                    xmlWriter.writeCharacters(persona.getDataNascita());
                    xmlWriter.writeEndElement();

                    xmlWriter.writeStartElement("comune_nascita");
                    xmlWriter.writeCharacters(persona.getComuneNascita());
                    xmlWriter.writeEndElement();

                    xmlWriter.writeStartElement("codice_fiscale");
                    xmlWriter.writeCharacters(persona.getCodiceFiscale());
                    xmlWriter.writeEndElement();

                    xmlWriter.writeStartElement("data_scadenza_id");
                    xmlWriter.writeCharacters(persona.getScadenzaDocumento());
                    xmlWriter.writeEndElement();

                    xmlWriter.writeStartElement("stato_transito");
                    xmlWriter.writeCharacters(persona.getStatoTransito());
                    xmlWriter.writeEndElement();

                    xmlWriter.writeEndElement(); // persona
                }

                xmlWriter.writeEndElement(); // persone
                xmlWriter.writeEndElement(); // giornata
            }

            xmlWriter.writeEndElement(); // diario
            xmlWriter.writeEndDocument();
            xmlWriter.flush();
        } catch (XMLStreamException | IOException e) {
            System.out.println("Error writing the file: " + e.getMessage());
        }
    }
}
